(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

open Atom
open Kinds

(* Once well-kindedness has been verified, all locations and kinds
   are dropped. *)

type constant = Types.constant =
  | TyUnit
  | TyArrow
  | TyPair
  | TyForall of kind
  | TyExists of kind

type ty =
  | TyVar of atom
  | TyConstant of constant
  | TyAbs of atom * ty
  | TyApp of ty * ty

type signature =
    ty AtomMap.t

(* ------------------------------------------------------------------------- *)

(* Equality between constants. *)

let constant_equality c1 c2 =
  match c1, c2 with
  | TyUnit, TyUnit
  | TyArrow, TyArrow
  | TyPair, TyPair ->
      true
  | TyForall kind1, TyForall kind2
  | TyExists kind1, TyExists kind2 ->
      equal kind1 kind2
  | _, _ ->
      false

(* ------------------------------------------------------------------------- *)

(* Converting types to raw types. *)

let rec erase (_, ty) =
  match ty with
  | Types.TyVar a ->
      TyVar a
  | Types.TyConstant c ->
      TyConstant c
  | Types.TyAbs (a, _, ty) ->
      TyAbs (a, erase ty)
  | Types.TyApp (ty1, ty2) ->
      TyApp (erase ty1, erase ty2)
  | Types.TyConstraint (ty, _) ->
      erase ty

(* ------------------------------------------------------------------------- *)

(* Display. *)

module P = struct

  open Pprint
  open Export

  let unit =
    text "()"

  let arrow =
    text " ->" ^^ softline

  let lambda =
    text "\\"

  let dot =
    text "." ^^ softline

  let forall =
    text "forall "

  let exists =
    text "exists "

  let comma =
    text "," ^^ softline

  (* [jump body] either displays [body] on a single line; or breaks a line,
     prints [body] at indentation 2, and breaks another line. *)

  let jump body =
    group (nest 2 (linebreak ^^ body) ^^ linebreak)

  let var env a =
    text (Identifier.name (resolve env a))

  let rec ty0 env = function
    | TyVar a ->
	var env a
    | TyConstant TyUnit ->
	unit
      (* the following constants should not appear isolated in the types that
	 are normally displayed to the user, but could appear isolated in
	 debugging output, e.g.  if the semi-algorithm is run in tracing
	 mode *)
    | TyConstant TyArrow ->
	text "(->)" (* TEMPORARY not parseable *)
    | TyConstant TyPair ->
	text "(,)" (* TEMPORARY not parseable *)
    | TyConstant (TyForall kind) ->
	forall ^^ brackets (text (print_kind kind))
    | TyConstant (TyExists kind) ->
	exists ^^ brackets (text (print_kind kind))
    | TyApp (TyApp (TyConstant TyPair, t1), t2) ->
	parens (
	  jump (
	    ty env t1 ^^ comma ^^ ty env t2
	  )
	)
    | t ->
	parens (ty env t)

  and ty1 env = function
    | TyApp (TyApp (TyConstant TyPair, _), _) as t ->
	ty0 env t
    | TyApp (t1, t2) ->
	ty1 env t1 ^^ space ^^ ty0 env t2
    | t ->
	ty0 env t

  and ty2 env t =
    ty1 env t (* no way of re-inserting backquotes *)

  and ty3 env = function
    | TyApp (TyApp (TyConstant TyArrow, t1), t2) ->
	ty2 env t1 ^^ arrow ^^ ty3 env t2
    | TyAbs (a, t) ->
	let env = bind env a in
	lambda ^^ var env a ^^ dot ^^ ty3 env t
    | TyApp (TyConstant (TyForall _), TyAbs (a, t)) -> (* TEMPORARY print kind too? *)
	let env = bind env a in
	forall ^^ var env a ^^ dot ^^ ty3 env t
    | TyApp (TyConstant (TyExists _), TyAbs (a, t)) ->
	let env = bind env a in
	exists ^^ var env a ^^ dot ^^ ty3 env t
    | t ->
	ty2 env t

  and ty env t =
    ty3 env t

  let ty : env * ty -> string =
    wrap (fun (env, t) -> ty env t)

end

let print_ty env t =
  P.ty (env, t)

(* ------------------------------------------------------------------------- *)

(* Copying. *)

(* Copying produces a fresh copy of a type. The copy is logically equal to the
   original type, that is, it represents the same type. The copy is fresh in
   the sense that its bound atoms are new. *)

(* The code is optimized by delaying the creation of a [phase] until we are
   forced to do so, that is, until we hit an abstraction. This optimizes the
   common case where the type contains no abstractions. *)

let rec icopy phase ty =
  match ty with
  | TyVar a ->
      let a' = copy_occurrence phase a in
      if a' == a then
	ty
      else
	TyVar a'
  | TyConstant _ ->
      ty
  | TyAbs (a, body) ->
      let a' = copy_binder phase a in
      TyAbs (a', icopy phase body)
  | TyApp (ty1, ty2) ->
      let ty1' = icopy phase ty1
      and ty2' = icopy phase ty2 in
      if ty1' == ty1 && ty2' == ty2 then
	ty
      else
	TyApp (ty1', ty2')

let rec copy ty =
  match ty with
  | TyVar _
  | TyConstant _ ->
      ty
  | TyAbs (a, body) ->
      let phase = setup() in
      let a' = copy_binder phase a in
      let ty' = TyAbs (a', icopy phase body) in
      close phase;
      ty'
  | TyApp (ty1, ty2) ->
      let ty1' = copy ty1
      and ty2' = copy ty2 in
      if ty1' == ty1 && ty2' == ty2 then
	ty
      else
	TyApp (ty1', ty2')

(* ------------------------------------------------------------------------- *)

(* Type equality (up to alpha-conversion). *)

let rec iequal phase ty1 ty2 =
  match ty1, ty2 with
  | TyVar a1, TyVar a2 ->
      map_occurrence phase a1 a2
  | TyConstant c1, TyConstant c2 ->
      if not (constant_equality c1 c2) then raise Map
  | TyAbs (a1, ty1), TyAbs (a2, ty2) ->
      map_binder phase a1 a2;
      iequal phase ty1 ty2
  | TyApp (ty11, ty12), TyApp (ty21, ty22) ->
      iequal phase ty11 ty21;
      iequal phase ty12 ty22
  | _, _ ->
      raise Map

let equal ty1 ty2 =
  let phase = setup() in
  try
    iequal phase ty1 ty2;
    close phase;
    true
  with Map ->
    close phase;
    false

(* ------------------------------------------------------------------------- *)

(* Checking whether a certain atom occurs free in a type. *)

let rec fresh_for a = function
  | TyVar b ->
      not (Atom.equal a b)
  | TyConstant _ ->
      true
  | TyAbs (b, ty) ->
      assert (not (Atom.equal a b)); (* binders should be distinct *)
      fresh_for a ty
  | TyApp (ty1, ty2) ->
      fresh_for a ty1 && fresh_for a ty2

(* ------------------------------------------------------------------------- *)

(* Listing the free atoms of a type. *)

(* The atoms that are defined at toplevel (that is, those that appear in
   the signature [types]) are excluded from the list. *)

let rec ifv types phase ty accu =
  match ty with
  | TyVar a ->
      if AtomMap.mem a types then
	accu
      else
	fv_occurrence phase a accu
  | TyConstant _ ->
      accu
  | TyAbs (a, ty) ->
      fv_binder phase a;
      ifv types phase ty accu
  | TyApp (ty1, ty2) ->
      let accu = ifv types phase ty1 accu in
      let accu = ifv types phase ty2 accu in
      accu

let fv types ty : atom list =
  let phase = setup() in
  let accu = ifv types phase ty [] in
  close phase;
  accu

(* ------------------------------------------------------------------------- *)

(* Capture-avoiding substitution. *)

(* I follow the (unpublished!) Linear FreshML discipline, that is, I maintain
   the invariant that (within a single term?) distinct binding sites bind
   distinct atoms. This offers the advantage that no freshening is required
   when traversing an abstraction.  On the other hand, the term [t] must be
   copied wherever an occurrence of the atom [a] is found. The complexity of
   the operation is linear in the size of the result. *)

let rec subst a t u =
  match u with
  | TyVar b ->
      if Atom.equal a b then copy t else u
  | TyConstant _ ->
      u
  | TyAbs (b, v) ->
      let v' = subst a t v in
      if v' == v then
	u
      else
	TyAbs (b, v')
  | TyApp (ty1, ty2) ->
      let ty1' = subst a t ty1
      and ty2' = subst a t ty2 in
      if ty1' == ty1 && ty2' == ty2 then
	u
      else
	TyApp (ty1', ty2')

(* ------------------------------------------------------------------------- *)

(* Reduction. *)

(* Reduction of a well-kinded type to a head normal form is guaranteed to
   terminate, even if toplevel type definitions are unfolded at will. *)

(* [hnf true] computes a weak head normal form, while [hnf false] computes
   a head normal form. If the argument [ty] is already in the desired form,
   then [ty] is returned, without any memory allocation. *)

(* TEMPORARY control which toplevel definitions to unfold? *)

let rec hnf types (weak : bool) ty =
  match ty with

  | TyVar a ->

      (* If this atom is defined at the toplevel, expand its definition.
	 Otherwise, this is a variable, hence a head normal form. *)

      if AtomMap.mem a types then
	hnf types weak (copy (AtomMap.find a types))
      else
	ty

  | TyConstant _ ->

      (* Constants do not reduce. *)

      ty

  | TyAbs (a, body) ->

      (* Reduction to head normal form takes place under abstractions;
	 reduction to weak head normal form does not. *)

      if weak then
	ty
      else
	let body' = hnf types weak body in
	if body' == body then
	  ty
	else
	  TyAbs (a, body')

  | TyApp (ty1, ty2) ->

      (* Reduce the left-hand side to weak head normal form. If an abstraction
	 appears, beta-reduce and continue; otherwise, we are done. *)
      
      begin match hnf types true ty1 with
      | TyAbs (a, body) ->
	  hnf types weak (subst a ty2 body)
      | ty1' ->
	  if ty1' == ty1 then
	    ty
	  else
	    TyApp (ty1', ty2)
      end

let hnf types ty =
  hnf types false ty

(* ------------------------------------------------------------------------- *)
(* ------------------------------------------------------------------------- *)

(* A semi-algorithm for determining whether two types are equal, in the sense
   that their Bhm trees coincide. *)

(* The semi-algorithm reports one of the following three outcomes. *)

type outcome =
  | True
  | Unknown
  | False

(* Internally, the semi-algorithm does not use the type [outcome] above.
   Instead, it uses the value [()] to mean [Unknown], and the exceptions
   [Valid] and [Invalid] to mean [True] and [False], respectively. *)

exception Valid

exception Invalid

let valid =
  Valid

let invalid =
  Invalid

(* A toggle for the debugging mode. *)

let debug =
  ref false

(* ------------------------------------------------------------------------- *)

(* Goal memoization. *)

(* The goals that we memoize are not necessarily in head normal form. Quite on
   the contrary, it would not make sense to insist that they be in head normal
   form, since a goal in head normal form is just a conjunction of subgoals
   that are not, in general, in head normal form; and it seems more reasonable
   to record individual subgoals, rather than conjunctions thereof. *)

(* It is valid for a goal to have free variables. However, when comparing two
   goals, we wish to compare them modulo renaming of their free variables. (Of
   course, the free variables of a goal must not include types that have been
   defined at the toplevel.) The easiest way of achieving this is to store closed
   goals, that is, goals that have been abstracted with respect to their free
   variables, and to compare closed goals up to alpha-equivalence. In short,
   we reconstruct (minimal, canonical) prefixes of universal quantifiers. *)

(* It would be nice if we could recognize that a goal is a substitution
   instance of a previous goal. However, this would require higher-order
   matching, in a setting where the terms are not even simply-typed. That
   does not sound easy, so, for the time being, we stick with matching up
   to alpha. *)

(* A closed goal consists of a quantifier prefix and a pair of types. The
   length of the quantifier prefix is recorded, as this allows a slight speed
   improvement in [coincide]. *)

type closed_goal =
    int * atom list * ty * ty

(* [abstract] turns a goal into a closed goal. *)

let abstract types ty1 ty2 : closed_goal =

  (* Collect the quantifiers. *)

  let phase = setup() in
  let quantifiers = ifv types phase ty1 (ifv types phase ty2 []) in
  close phase;

  (* Copy the abstract goal [(quantifiers, ty1, ty2)], so there is no
     sharing between the history and the current goal. *)

  let phase = setup() in
  let quantifiers = List.map (copy_binder phase) quantifiers in
  let ty1 = icopy phase ty1
  and ty2 = icopy phase ty2 in
  close phase;

  (* Done. *)

  (List.length quantifiers, quantifiers, ty1, ty2)

(* [coincide] determines whether two closed goals are equal, up to
   alpha-equivalence. *)

let coincide (nq1, quantifiers1, ty11, ty12) (nq2, quantifiers2, ty21, ty22) =
  nq1 = nq2 &&
    (* guarantees that [iter2] will not fail, and offers a quick path *)
  let phase = setup() in
  List.iter2 (map_binder phase) quantifiers1 quantifiers2;
  try
    iequal phase ty11 ty21;
    iequal phase ty12 ty22;
    close phase;
    true
  with Map ->
    close phase;
    false

(* The history is just a list of closed goals. *)

(* TEMPORARY perhaps one could do better, e.g. index them by the number of
   their quantifiers, or somehow set up a (persistent, alpha-compatible)
   hash table *)

type history =
    closed_goal list

(* The empty history. *)

let empty =
  []

(* [check ty1 ty2 h] checks whether the equation [ty1 = ty2] appears within
   the history [h]. If it does, then [Valid] is raised; otherwise, an extended
   history is returned. *)

let check types ty1 ty2 h =
  let goal = abstract types ty1 ty2 in
  if List.exists (coincide goal) h then
    raise valid
  else
    (* Construct the symmetric goal. We must invoke [abstract] again, as the
       canonical order of the quantifiers might be different. *)
    let symmetric_goal = abstract types ty2 ty1 in
    if List.exists (coincide symmetric_goal) h then
      raise valid
    else
      goal :: h

(* [insert types ty1 ty2 h] extends the history [h] with the equation [ty1 = ty2]. *)

let insert types ty1 ty2 h =
  try
    check types ty1 ty2 h
  with Valid ->
    h

(* ------------------------------------------------------------------------- *)

(* State. *)

(* The state of the semi-algorithm, besides the current goal, consists of the
   following record. *)

type state = {

  (* The toplevel signature. This signature does not change during a
     run of the semi-algorithm. *)

  types: signature;

  (* An export environment. This is used only for printing traces, and
     could be removed if desired. *)

  xenv: Export.env;

  (* A depth bound. This value is decremented as the semi-algorithm
     goes down a branch; the branch is pruned when it reaches zero. *)

  depth: int;

  (* A history (see above). *)

  history: history;

}

(* ------------------------------------------------------------------------- *)

(* The semi-algorithm is organized as a depth-first search algorithm, because
   this is easy to understand and efficient. In order to ensure termination,
   the search depth is bounded. *)

(* The algorithm is not purely conjunctive -- it also contains a heuristic
   component, which causes it to explore disjunctions. The presence of
   disjunctions makes it desirable to quickly refute invalid subgoals. To this
   end, iterative deepening is used. (It is not clear how effective this
   really is. In particular, iterative deepening is used at the toplevel
   only.) *)

(* Iterative deepening. *)

(* The parameter [depth] represents the maximum depth that we are willing to
   search to, whereas [s.depth] represents the depth bound of the upcoming
   iteration. *)

let rec iterate depth (search : state -> unit) s =
  if s.depth <= depth then begin
    (* Search at depth [s.depth]. *)
    search s;
    (* If inconclusive, try again at a greater depth. *)
    iterate depth search { s with depth = s.depth + 1 }
  end

(* ------------------------------------------------------------------------- *)

(* Mini-helpers. *)

(* This extends the export environment. *)

let bind s a =
  { s with xenv = Export.bind s.xenv a }

(* This prints a goal. *)

let print_goal s ty1 ty2 =
  Printf.printf "%s = %s\n%!" (print_ty s.xenv ty1) (print_ty s.xenv ty2)

(* ------------------------------------------------------------------------- *)

(* Refutation. *)

(* We begin with a simplified version of the semi-algorithm, which is designed
   to refute goals. *)

(* [refute] implements depth-first search. It returns unit if the search is
   inconclusive (due to pruning) or if the goal is valid, and raises [Invalid]
   if the goal can be proved invalid. It requires a goal in head normal
   form. This algorithm is purely conjunctive (it does not explore any
   disjunctions), so the exception [Invalid] is never caught internally.
   Because we do not attempt to prove the goal, a history is not useful. *)

let rec refute ty1 ty2 (s : state) : unit =
  match ty1, ty2 with

  | TyAbs (a1, ty1), TyAbs (a2, ty2) ->
      
      (* Descending under a pair of lambda-abstractions requires renaming. *)

      let a = fresha a1 in
      refute
	(subst a1 (TyVar a) ty1)
	(subst a2 (TyVar a) ty2)
	(bind s a)

  | TyAbs (a1, ty1), ty2 ->

      (* If there is an abstraction on one side only, then the other side must
	 be eta-expanded. Generating a fresh atom is not required, as [a1]
	 cannot occur in [ty2]. *)

      refute
	ty1
	(TyApp (ty2, TyVar a1))
	(bind s a1)

  | ty1, TyAbs (a2, ty2) ->

      refute
	(TyApp (ty1, TyVar a2))
	ty2
	(bind s a2)

  | TyApp (ty11, ty12), TyApp (ty21, ty22) ->

      (* Because this are head normal forms, applications act as conjunctions,
	 and must be decomposed. The left-hand side, which is already in head
	 normal form, is recursively decomposed. The right-hand side is not in
	 head normal form, so it must be normalized. *)

      (* The right-hand side lies one level deeper in the Bhm tree, so
	 [s.depth] is decreased. If it was zero already, the search stops,
	 with an inconclusive result -- we just return unit. *)

      refute ty11 ty21 s;

      if s.depth > 0 then
	refute_aux ty12 ty22 { s with depth = s.depth - 1 }

  | TyVar a1, TyVar a2 ->

      if Atom.equal a1 a2 then () else raise invalid

  | TyConstant c1, TyConstant c2 ->

      if constant_equality c1 c2 then () else raise invalid

  | TyVar _, TyConstant _
  | TyConstant _, TyVar _
  | (TyVar _ | TyConstant _), TyApp _
  | TyApp _, (TyVar _ | TyConstant _) ->

      raise invalid

(* [refute_aux] is analogous to [refute], but does not make the assumption
   that [ty1] and [ty2] are head normal forms. *)

and refute_aux ty1 ty2 (s : state) : unit =

  (* If the subgoal happens to be a tautology, it is removed. This check
     can lead to termination in cases that would otherwise diverge. We
     check this before reduction (as a shortcut) and after reduction. *)

  if equal ty1 ty2 then
    ()
  else
    let ty1n = hnf s.types ty1
    and ty2n = hnf s.types ty2 in
    if equal ty1n ty2n then
      ()
    else

      (* Normalize the goal and hand it over to [refute]. *)
      
      refute ty1n ty2n s

(* ------------------------------------------------------------------------- *)

(* Goal decomposition. *)

(* The following two recursive functions implement the depth-first
   search. They return unit if the search is inconclusive (due to pruning),
   and raise [Valid] or [Invalid] if a definite result is obtained. *)

let enable_heuristic =
  ref true

(* [decompose] decomposes a goal that is already in head normal form. This
   yields a conjunction of subgoals that are not necessarily in head normal
   form. These subgoals are handled by [decompose_aux]. *)

let rec decompose ty1 ty2 (s : state) : unit =
  match ty1, ty2 with

  | TyAbs (a1, ty1), TyAbs (a2, ty2) ->
      
      (* Descending under a pair of lambda-abstractions requires renaming. *)

      let a = fresha a1 in
      decompose
	(subst a1 (TyVar a) ty1)
	(subst a2 (TyVar a) ty2)
	(bind s a)

  | TyAbs (a1, ty1), ty2 ->

      (* If there is an abstraction on one side only, then the other side must
	 be eta-expanded. Generating a fresh atom is not required, as [a1]
	 cannot occur in [ty2]. *)

      decompose
	ty1
	(TyApp (ty2, TyVar a1))
	(bind s a1)

  | ty1, TyAbs (a2, ty2) ->

      decompose
	(TyApp (ty1, TyVar a2))
	ty2
	(bind s a2)

  | TyApp (ty11, ty12), TyApp (ty21, ty22) ->

      (* Because this are head normal forms, applications act as conjunctions,
	 and must be decomposed. The left-hand side, which is already in head
	 normal form, is recursively decomposed. The right-hand side is not in
	 head normal form, so it is passed to [decompose_aux]. *)

      begin try
	  decompose ty11 ty21 s
	    (* If this call returns, then the truth value of the left-hand
	       side is unknown; there is little point, then, in evaluating the
	       right-hand side. One could evaluate the right-hand side in the
	       hope of proving it invalid, but that might add to our runtime. *)
      with Valid ->
	(* The left-hand side is valid; examine the right-hand side. *)
	(* The right-hand side lies one level deeper in the Bhm tree,
	   so [s.depth] is decreased. If it was zero already, the search
	   stops, with an inconclusive result -- we just return unit. *)
	if s.depth > 0 then
	  decompose_aux ty12 ty22 { s with depth = s.depth - 1 }

	(* If the left-hand side is invalid, there is no point in evaluating
	   the right-hand side; we let the exception [Invalid] escape. *)
      end

  | TyVar a1, TyVar a2 ->

      raise (if Atom.equal a1 a2 then valid else invalid)

  | TyConstant c1, TyConstant c2 ->

      raise (if constant_equality c1 c2 then valid else invalid)

  | TyVar _, TyConstant _
  | TyConstant _, TyVar _
  | (TyVar _ | TyConstant _), TyApp _
  | TyApp _, (TyVar _ | TyConstant _) ->

      raise invalid

(* [decompose_aux] is analogous to [decompose], but does not make the assumption
   that [ty1] and [ty2] are head normal forms. In its simplest form, it would
   just reduce the goal to head normal form and invoke [decompose] again.
   However, this basic scheme can be improved in a few important ways. *)

and decompose_aux ty1 ty2 (s : state) : unit =
  match ty1, ty2 with

  | TyAbs (a1, ty1), TyAbs (a2, ty2) ->
      
      (* As in [decompose], if the goal exhibits an abstraction on each side,
	 we descend under it. *)

      let a = fresha a1 in
      decompose_aux
	(subst a1 (TyVar a) ty1)
	(subst a2 (TyVar a) ty2)
	(bind s a)

  | _, _ ->

    (* If the subgoal happens to be a tautology, it is removed. This check
       can lead to termination in cases that would otherwise diverge. We
       check this before reduction (as a shortcut) and after reduction. *)

    if equal ty1 ty2 then
      raise valid;

    let ty1n = hnf s.types ty1
    and ty2n = hnf s.types ty2 in
    if equal ty1n ty2n then
      raise valid;

    (* Check whether the goal occurs in the history. (As argued above, it makes
       sense to use the non-normalized goal here.) If it does, then we declare
       it valid ([check] raises [Valid]). Otherwise, we add the goal to the
       history, so that it is declared valid when it is next encountered. This
       makes sense because this next encounter can only take place deeper in the
       Bhm tree. (Indeed, the recursive calls to [decompose_aux] will not find the
       same goal again, because they work on syntactically smaller goals; it is
       only after [decompose_aux] calls [decompose] and [decompose] calls [decompose_aux]
       that this goal can be encountered again.) *)

    (* TEMPORARY the history grows only along one branch; should it also grow
       across branches, i.e., after a subgoal has been found valid, should it
       be added to the history that is used to examine its siblings? *)

    let s = { s with history = check s.types ty1 ty2 s.history } in

    (* Now, we shall make two attempts at determining the truth value of the
       goal. The first attempt is sound and complete, so it can prove or
       disprove the goal; while the second attempt is sound and incomplete, so
       it can prove, but not disprove, the goal. *)

    (* First attempt: normalize the goal and hand it over to [decompose]. If the
       goal is false, it is hoped that this approach might quickly refute it, so
       we will not waste any time in the second attempt. *)

    (* If this first attempt raises [Valid] or [Invalid], then we are happy,
       and let the exception escape. *)

    decompose ty1n ty2n s;

    (* If the above call returns, then our first attempt was indecisive: the
       truth value of the goal remains unknown. Instead of abandoning, we now
       turn to the second attempt. If the goal is an equation between two
       applications, we decompose this equation without first normalizing
       it. This might seem futile. However, if we are lucky, this can get some
       troublesome subterms out of the way, and actually lead to success. If
       this heuristic guess is doomed, this branch should die off quickly
       anyway, as (it is hoped that) the semi-algorithm is good at refuting
       invalid subgoals. *)

    (* Of course, this second attempt is made only if the goal is not in
       head normal form; otherwise it would just repeat the first attempt. *)

    (* TEMPORARY Before attempting to prove the goal, one could try to quickly
       refute it. This makes sense because we are using an incomplete
       heuristic (see below) that might produce nonsensical goals. We don't
       want to waste time on these goals, so a sanity check seems useful.
       However, I haven't been able to gain any time with this idea. *)

    if !enable_heuristic && (ty1 != ty1n || ty2 != ty2n) then
      match ty1, ty2 with
      | TyApp (ty11, ty12), TyApp (ty21, ty22) -> (

	  try
	    try
	      decompose_aux ty11 ty21 s
	    with Valid ->
	      decompose_aux ty12 ty22 s
	  with Invalid ->
	    (* Because this was only a heuristic guess, its failure is not
	       meaningful. *)
	    ()

	)
    | _, _ ->
	()

(* ------------------------------------------------------------------------- *)

(* The body of the semi-algorithm. *)

(* The maximum refutation depth. *)

(* This parameter cannot be very high: if the goal is true (a typical case),
   then the refutation algorithm will run down to the full depth. *)

let refutation_depth =
  16

(* The maximum search depth. *)

let search_depth =
  16

(* A global history is used to record statements that have been proved true
   and that make sense in the toplevel signature. (Typically, these are the
   lemmas stated by the user at the top level.) This helps speed up further
   decision problems. *)

let global_history =
  ref empty

(* The main entry point. *)

let semidecide global types xenv ty1 ty2 =

  (* Set up the initial state. *)

  let s = {
    types = types;
    xenv = xenv;
    depth = 0; (* the initial search depth *)
    history = !global_history;
  } in

  try

    (* We first try to refute the goal, because the refutation semi-algorithm
       is faster (it does not include any heuristics) and complete: if the
       goal is false, it is certain that it will be refuted, provided the
       mismatch lies at a reasonable depth. *)

    (* Because we expect the goal to be typically true, we do not use
       iterative deepening. *)

    refute_aux ty1 ty2 { s with depth = refutation_depth };

    (* If the refutation process is inconclusive, try to prove the goal. *)

    iterate search_depth (decompose_aux ty1 ty2) s;

    (* If both processes fail, report. *)

    Unknown

  with
  | Valid ->
      if global then
	global_history := insert types ty1 ty2 !global_history;
      True
  | Invalid ->
      False

(* ------------------------------------------------------------------------- *)
(* ------------------------------------------------------------------------- *)

(* TEMPORARY since the history is untyped, we might accidentally consider
   a goal valid, because it occurs in the history with a different kind.
   Is this a problem? (probably not) *)

