(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

open Syntax
open Atom
open Import

(* ------------------------------------------------------------------------- *)

(* [is_auxiliary] recognizes a set of so-called auxiliary characters: prime,
   underscore, digit. *)

let is_auxiliary = function
  | '\''
  | '_'
  | '0' .. '9' ->
      true
  | _ ->
      false

(* [starts_with s1 s2] checks whether [s1] is a prefix of [s2] and the remainder
   of [s2] is composed of auxiliary characters. *)

let starts_with s1 s2 =
  let n1 = String.length s1
  and n2 = String.length s2 in
  n1 <= n2 &&
  let rec loop1 i =
    (i >= n1 && loop2 i) ||
    (i < n1 && s1.[i] = s2.[i] && loop1 (i + 1))
  and loop2 i =
    i >= n2 ||
    (is_auxiliary s2.[i] && loop2 (i + 1))
  in
  loop1 0

(* [find s ss] looks for the key [s] within the association list [ss], using
   [starts_with] as a comparison function. *)

let rec find s2 = function
  | [] ->
      None
  | (s1, data) :: rest ->
      if starts_with s1 s2 then Some data else find s2 rest

(* [get_default sort defaults id] finds a default kind (or type, or ...) for the
   identifier [id] in the list [defaults]. It takes care of error reporting if
   the search fails. *)

let get_default sort defaults id =
  match find (Identifier.name id) defaults with
  | Some data ->
      data
  | None ->
      Error.error
	[ Identifier.location id ]
	(Printf.sprintf "No default %s was specified for \"%s\"." sort (Identifier.name id))

(* ------------------------------------------------------------------------- *)

(* Kinds. *)

let rec kind env = function
  | KStar ->
      Kinds.KStar
  | KDefined id ->
      Kinds.KDefined (resolve env id)
  | KArrow (k1, k2) ->
      Kinds.KArrow (kind env k1, kind env k2)
  | KLater k ->
      Kinds.KLater (kind env k)

(* ------------------------------------------------------------------------- *)

(* Types. *)

and optional_kind defaults env id = function
  | Some k ->
      kind env k
  | None ->
      get_default "kind" defaults id

and constant_of_quantifier q k =
  match q with
  | QForall ->
      Types.TyForall k
  | QExists ->
      Types.TyExists k

and ty loc defaults env = function
  | TyVar id ->
      Types.TyVar (resolve env id)
  | TyAbs (id, ko, t) ->
      let env = bind env id in
      Types.TyAbs (
	resolve env id,
	optional_kind defaults env id ko,
	locty defaults env t
      )
  | TyApp (ty1, ty2) ->
      Types.TyApp (locty defaults env ty1, locty defaults env ty2)
  | TyConstraint (t, k) ->
      Types.TyConstraint (locty defaults env t, kind env k)
  | TyUnit ->
      Types.TyConstant Types.TyUnit
  | TyArrow (loc1, t1, t2) ->
      let ty0 = Error.dummy, Types.TyConstant Types.TyArrow in
      let ty1 = loc1, Types.TyApp (ty0, locty defaults env t1) in
      Types.TyApp (ty1, locty defaults env t2)
  | TyPair (loc1, t1, t2) ->
      let ty0 = Error.dummy, Types.TyConstant Types.TyPair in
      let ty1 = loc1, Types.TyApp (ty0, locty defaults env t1) in
      Types.TyApp (ty1, locty defaults env t2)
  | TyQuantified (q, id, ko, t) ->
      let env = bind env id in
      let k = optional_kind defaults env id ko in
      let c = constant_of_quantifier q k in
      let ty0 = Error.dummy, Types.TyConstant c in
      let ty1 = loc, Types.TyAbs (
	resolve env id,
	k,
	locty defaults env t
      ) in
      Types.TyApp (ty0, ty1)
  | TyRawQuantifier (q, k) ->
      let k = kind env k in
      Types.TyConstant (constant_of_quantifier q k)

and locty defaults env (loc, t) =
  (loc, ty loc defaults env t)

(* ------------------------------------------------------------------------- *)

(* Statements. *)

and statement defaults env = function
  | SEquation (t1, t2) ->
      Lemmas.SEquation (locty defaults env t1, locty defaults env t2)
  | SForall (id, ko, s) ->
      let env = bind env id in
      Lemmas.SForall (
	resolve env id,
	optional_kind defaults env id ko,
	statement defaults env s
      )

(* ------------------------------------------------------------------------- *)

(* Terms. *)

and term loc defaults env = function
  | TeVar id ->
      Terms.TeVar (resolve env id)
  | TeAbs (id, ty, t) ->
      let env = bind env id in
      Terms.TeAbs (
	resolve env id,
	locty defaults env ty,
	locterm defaults env t
      )
  | TeApp (t1, t2) ->
      Terms.TeApp (
	locterm defaults env t1,
	locterm defaults env t2
      )
  | TeTyAbs (id, ko, t) ->
      let env = bind env id in
      Terms.TeTyAbs (
	resolve env id,
	optional_kind defaults env id ko,
	locterm defaults env t
      )
  | TeTyApp (t, ty) ->
      Terms.TeTyApp (
	locterm defaults env t,
	locty defaults env ty
      )
  | TePack (t, ty1, ty2) ->
      Terms.TePack (
	locterm defaults env t,
	locty defaults env ty1,
	locty defaults env ty2
      )
  | TeUnpack (alpha, x, t1, t2) ->
      let t1 = locterm defaults env t1 in
      let env = bind env alpha in
      let env = bind env x in
      Terms.TeUnpack (
	resolve env alpha,
	resolve env x,
	t1,
	locterm defaults env t2
      )
  | TeConstraint (t, ty) ->
      Terms.TeConstraint (
	locterm defaults env t,
	locty defaults env ty
      )
  | TeType (id, ty, t) ->
      let ty = locty defaults env ty in
      let env = bind env id in
      Terms.TeType (
	resolve env id,
	ty,
	locterm defaults env t
      )
  | TeLet (id, t1, t2) ->
      let t1 = locterm defaults env t1 in
      let env = bind env id in
      Terms.TeLet (
	resolve env id,
	t1,
	locterm defaults env t2
      )
  | TeUnit ->
      Terms.TeUnit
  | TePair (t1, t2) ->
      Terms.TePair (
	locterm defaults env t1,
	locterm defaults env t2
      )
  | TeLetPair (id1, id2, t1, t2) ->
      let t1 = locterm defaults env t1 in
      let env = bind_simultaneously env (
	match id1, id2 with
	| Some id1, Some id2 ->
	    [ id1; id2 ]
	| Some id, None
	| None, Some id ->
	    [ id ]
	| None, None ->
	    []
      ) in
      Terms.TeLetPair (
	Option.map (resolve env) id1,
	Option.map (resolve env) id2,
	t1,
	locterm defaults env t2
      )

and locterm defaults env (loc, t) =
  (loc, term loc defaults env t)

(* ------------------------------------------------------------------------- *)

(* Declarations. *)

(* Gathering bound variables (of all sorts). *)

and bvar accu = function
  | DKind (id, _)
  | DType (id, _, _) 
  | DTerm (id, _, _) ->
      id :: accu
  | DDefaultKind _
  | DLemma _ ->
      accu

and bvars ds =
  List.fold_left bvar [] ds

(* Gathering default kind declarations. *)

and gather_defaults env d defaults =
  match d with
  | DDefaultKind (id, k) ->
      (Identifier.name id, kind env k) :: defaults
  | DKind _
  | DType _
  | DLemma _
  | DTerm _ ->
      defaults

and declaration defaults env (kinds, types, lemmas, terms) = function
  | DKind (id, k) ->
      AtomMap.add (resolve env id) (kind env k) kinds,
      types,
      lemmas,
      terms
  | DType (id, k, t) ->
      kinds,
      AtomMap.add (resolve env id) (kind env k, locty defaults env t) types,
      lemmas,
      terms
  | DDefaultKind _ ->
      kinds,
      types,
      lemmas,
      terms
  | DLemma s ->
      kinds,
      types,
      statement defaults env s :: lemmas,
      terms
  | DTerm (id, ty, t) ->
      kinds,
      types,
      lemmas,
      AtomMap.add (resolve env id) (locty defaults env ty, locterm defaults env t) terms

and declarations ds =
  let env =
    bind_simultaneously empty (bvars ds)
  in
  let defaults =
    List.fold_right (gather_defaults env) ds []
  in
  List.fold_left (declaration defaults env) (AtomMap.empty, AtomMap.empty, [], AtomMap.empty) ds

(* ------------------------------------------------------------------------- *)

(* Programs. *)

let program p =
  let kinds, types, lemmas, terms = declarations p in
  Error.signaled();
  Kinds.install kinds;
  types, List.rev lemmas, terms

