(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

open Printf
open Atom
open Kinds
open Types

(* The terms in FORK are the same as in System F or F^omega. *)

type term =
  | TeVar of atom
      (* x *)
  | TeAbs of atom * locty * locterm
      (* \x : ty. t *)
  | TeApp of locterm * locterm
      (* t t *)
  | TeTyAbs of atom * kind * locterm
      (* /\ a : k. t *)
  | TeTyApp of locterm * locty
      (* t [ ty ] *)
  | TePack of locterm * (* existential goal *) locty * (* witness *) locty
      (* pack ty, t as ty *)
  | TeUnpack of atom * atom * locterm * locterm
      (* unpack a, x = t in t *)
  | TeConstraint of locterm * locty
      (* (t : ty) *)
  | TeType of atom * locty * locterm
      (* type a = ty in t *)
  | TeLet of atom * locterm * locterm
      (* let x = t in t *)
  | TeUnit
      (* () *)
  | TePair of locterm * locterm
      (* (t, t) *)
  | TeLetPair of atom option * atom option * locterm * locterm
      (* let (x, x) = t in t *)

and locterm =
    Error.location * term

(* Terms can be (recursively) defined by the user. A term signature
   maps atoms to terms. *)

type signature =
    (locty * locterm) AtomMap.t

(* ------------------------------------------------------------------------- *)

(* Typechecking. *)

exception TypeChecking of Error.location list * string

(* ------------------------------------------------------------------------- *)

(* What is the kind of the types that classify values? In System F^omega, it
   is the base kind [KStar]. Here, however, we might need to also allow
   iterated applications of [KLater] to [KStar]. *)

(* [has_base_kind kenv ty] checks that [ty] is well-kinded and has a base kind.
   It returns [erase ty]. *)

open RawTypes

let has_base_kind kenv ((loc, _) as ty) =
  let kind = Types.infer kenv ty in
  if is_later_star kind then
    erase ty
  else
    raise (TypeChecking ([loc], sprintf
      "This type should have kind * (or later). Instead, it has kind:\n\
         %s"
      (print_kind kind)
    ))

(* ------------------------------------------------------------------------- *)

(* One beta-reduction step sometimes helps produce a more readable type. *)

let app ty1 ty2 =
  match ty1 with
  | TyAbs (a1, body1) ->
      subst a1 ty2 body1
  | _ ->
      TyApp (ty1, ty2)

(* ------------------------------------------------------------------------- *)

(* Synthesizing the type of a term is a bottom-up process. *)

let rec infer types xenv kenv tenv (loc, t) =
  match t with
  | TeVar x ->
      copy (AtomMap.find x tenv)
      (* copying is mandatory to preserve the Linear FreshML invariant:
	 binders are unique! *)
  | TeAbs (x, ty, t) ->
      let ty = has_base_kind kenv ty in
      let xenv = Export.bind xenv x in
      let tenv = AtomMap.add x ty tenv in
      TyApp (TyApp (TyConstant TyArrow, ty), infer types xenv kenv tenv t)
  | TeApp (t1, t2) ->
      let ty1 = infer types xenv kenv tenv t1
      and ty2 = infer types xenv kenv tenv t2 in
      begin match hnf types ty1 with
      | TyApp (TyApp (TyConstant TyArrow, ty11), ty12) ->
	  begin match semidecide false types xenv ty2 ty11 with
	  | True ->
	      ty12
	  | Unknown ->
	      raise (TypeChecking ([loc], sprintf
		"This application might be invalid. The actual argument has type:\n\
		 %s\n\
		 while the formal argument has type:\n\
		 %s\n\
                 I was unable to prove or disprove this type equality."
		(print_ty xenv ty2)
		(print_ty xenv ty11)
	      ))
	  | False ->
	      raise (TypeChecking ([loc], sprintf
		"This application is invalid. The actual argument has type:\n\
		 %s\n\
		 while the formal argument has type:\n\
		 %s"
		(print_ty xenv ty2)
		(print_ty xenv ty11)
	      ))
	  end
      | ty1 ->
	  raise (TypeChecking ([loc], sprintf
	    "This application is invalid. The subject is not a function.\n\
	     It has type:\n\
	     %s"
	    (print_ty xenv ty1)
	  ))
      end
  | TeTyAbs (a, kind, t) ->
      let xenv = Export.bind xenv a in
      let kenv = AtomMap.add a kind kenv in
      TyApp (TyConstant (TyForall kind), TyAbs (a, infer types xenv kenv tenv t))
      (* TEMPORARY qui garantit que ce type est bien sort? *)
      (* je crois qu'il faut une constante forall polymorphe en n *)
  | TeTyApp (t, ty2) ->
      let ty1 = infer types xenv kenv tenv t in
      let kind2 = Types.infer kenv ty2 in
      begin match hnf types ty1 with
      | TyApp (TyConstant (TyForall kind1), body1) ->
	  if is_subkind_later kind2 kind1 then
	    app body1 (erase ty2)
	  else
	    raise (TypeChecking ([loc], sprintf
	      "This type application is invalid. The actual argument has kind:\n\
               %s\n\
               while the formal argument has kind:\n\
               %s"
	      (print_kind kind2)
	      (print_kind kind1)
	    ))
      | ty1 ->
	  raise (TypeChecking ([loc], sprintf
	    "This type application is invalid. The subject is not a type function.\n\
             It has type:\n\
             %s"
	    (print_ty xenv ty1)
	  ))
      end
  | TePack (t, ((xloc, _) as existential), witness) ->
      let existential = has_base_kind kenv existential in
      let kind2 = Types.infer kenv witness in
      begin match hnf types existential with
      | TyApp (TyConstant (TyExists kind1), body1) ->
	  if is_subkind_later kind2 kind1 then begin
	    let (_ : ty) = raw_check types xenv kenv tenv loc t (app body1 (erase witness)) in
	    existential
	  end
	  else
	    raise (TypeChecking ([loc], sprintf
	      "This use of pack is invalid. The witness has kind:\n\
               %s\n\
               while the existential quantification is at kind:\n\
               %s"
	      (print_kind kind2)
	      (print_kind kind1)
	    ))
      | existential ->
	  raise (TypeChecking ([xloc], sprintf
	    "This use of pack is invalid. This is not an existential type.\n\
             Its head normal form is:\n\
             %s"
	    (print_ty xenv existential)
	  ))
      end
  | TeUnpack (a, x, t1, t2) ->
      let ty1 = infer types xenv kenv tenv t1 in
      begin match hnf types ty1 with
      | TyApp (TyConstant (TyExists kind1), body1) ->
	  let xenv = Export.bind xenv a in
	  let kenv = AtomMap.add a kind1 kenv in
	  let xenv = Export.bind xenv x in
	  let tenv = AtomMap.add x (app body1 (TyVar a)) tenv in
	  let ty2 = infer types xenv kenv tenv t2 in
	  (* TEMPORARY maybe one should reduce ty2 prior to checking whether
	     a occurs in it; reduction could cause occurrences of a to vanish *)
	  (* change the error message: "seems to escape" / "seems to occur free" ? *)
	  if fresh_for a ty2 then
	    ty2
	  else
	    raise (TypeChecking ([loc], sprintf
	      "The type variable %s escapes its scope. It occurs free in the type:\n\
               %s"
	      (print_ty xenv (TyVar a))
	      (print_ty xenv ty2)
	    ))
      | ty1 ->
	  raise (TypeChecking ([loc], sprintf
	    "This use of unpack is invalid. The subject is not an existential package.\n\
	     It has type:\n\
	     %s"
	    (print_ty xenv ty1)
	  ))	  
      end
  | TeLet (x, t1, t2) ->
      let ty1 = infer types xenv kenv tenv t1 in
      let xenv = Export.bind xenv x in
      let tenv = AtomMap.add x ty1 tenv in
      infer types xenv kenv tenv t2
  | TeConstraint (t, ty) ->
      check types xenv kenv tenv loc t ty
  | TeType (a, ty, t) ->
      (* for simplicity, local type definitions are considered non-recursive.
	 this allows us to easily expand them away when we exit their scope. *)
      let kind = Types.infer kenv ty in
      let ty = erase ty in
      let xenv = Export.bind xenv a in
      let kenv = AtomMap.add a kind kenv in
      let types = AtomMap.add a ty types in
      let body = infer types xenv kenv tenv t in
      subst a ty body (* expand [a] away *)
  | TeUnit ->
      TyConstant TyUnit
  | TePair (t1, t2) ->
      let ty1 = infer types xenv kenv tenv t1
      and ty2 = infer types xenv kenv tenv t2 in
      TyApp (TyApp (TyConstant TyPair, ty1), ty2)
  | TeLetPair (x, y, t1, t2) ->
      let ty1 = infer types xenv kenv tenv t1 in
      begin match hnf types ty1 with
      | TyApp (TyApp (TyConstant TyPair, ty11), ty12) ->
	  let xenv, tenv =
	    match x with
	    | Some x ->
		Export.bind xenv x,
		AtomMap.add x ty11 tenv
	    | None ->
		xenv, tenv
	  in
	  let xenv, tenv =
	    match y with
	    | Some y ->
		Export.bind xenv y,
		AtomMap.add y ty12 tenv
	    | None ->
		xenv, tenv
	  in
	  infer types xenv kenv tenv t2
      | ty1 ->
	  raise (TypeChecking ([loc], sprintf
	    "This pair deconstruction is invalid. The subject is not a pair.\n\
	     It has type:\n\
	     %s"
	    (print_ty xenv ty1)
	  ))	  
      end

and check types xenv kenv tenv loc t declared_type =
  let declared_type = has_base_kind kenv declared_type in
  raw_check types xenv kenv tenv loc t declared_type

and raw_check types xenv kenv tenv loc t declared_type =
  let inferred_type = infer types xenv kenv tenv t in
  match semidecide false types xenv inferred_type declared_type with
  | True ->
      declared_type
  | Unknown ->
      raise (TypeChecking ([loc], sprintf
	"This type annotation might be invalid.\n\
         This term has type:\n\
	 %s\n\
         whereas the expected type is:\n\
	 %s\n\
	 I was unable to prove or disprove this type equality."
	(print_ty xenv inferred_type)
	(print_ty xenv declared_type)
      ))
  | False ->
      raise (TypeChecking ([loc], sprintf
	"This type annotation is invalid.\n\
         This term has type:\n\
	 %s\n\
         whereas the expected type is:\n\
	 %s"
	(print_ty xenv inferred_type)
	(print_ty xenv declared_type)
      ))

(* ------------------------------------------------------------------------- *)

(* Typechecking the toplevel term definitions. *)

(* The term definitions found at the top level are considered recursive. For
   this reason, they are annotated with a type. Checking these definitions
   involves matching the inferred type against the declared type. *)

let check kinds types xenv terms =
  let tenv =
    try
      AtomMap.map (fun (declared_type, _) -> has_base_kind kinds declared_type) terms
    with
    | KindChecking (locs, msg)
    | TypeChecking (locs, msg) ->
	Error.error locs msg
  in
  AtomMap.iter (fun x (declared_type, t) ->
    try
      let loc, _ = t in
      let (_ : ty) = check types xenv kinds tenv loc t declared_type in
      ()
    with
    | KindChecking (locs, msg)
    | TypeChecking (locs, msg) ->
	Error.signal locs msg
  ) terms;
  Error.signaled();
  tenv

